/*
 * Copyright (c) 2023. Helge Eichhorn and the LOX contributors
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, you can obtain one at https://mozilla.org/MPL/2.0/.
 */

//! Module `python` aggregates the Python binding for `lox-time`.

pub mod deltas;
pub mod time;
pub mod time_scales;
pub mod ut1;
pub mod utc;
